unit Element;

{$IMPORTEDDATA OFF}

interface

uses NTypes, NGraphics;

type
  TCurrent = record
    Y: Integer;
    Color: Integer;
  end;

  TElement = class
    Args: TStringArray;
    Values: TStringArray;
    Height: Integer;
    Cond: string;
    CondValue: Boolean;
    constructor Create(const Cond: string; const Args: TStringArray); virtual;
    procedure Draw(const Canvas: TCanvas); virtual;
    procedure DrawAlpha(const Canvas: TCanvas);
    procedure Init; virtual;
    procedure Update; virtual;
  end;

var
  Width: Integer;
  Height: Integer;
  BgAlpha: Integer;
  Current: TCurrent;

function CreateElement(const Args: TStringArray): TElement;
procedure ClearFontCache;

implementation

uses NLib, NSys, Monitor;

type
  TElementClass = class of TElement;

  TGraphElement = class(TElement)
    Cur: Integer;
    Data: array of Double;
    Range: Double;
    Prev: Double;
    procedure Update; override;
  end;

  TEareagraph = class(TGraphElement)
    procedure Draw(const Canvas: TCanvas); override;
  end;

  TEbargraph = class(TGraphElement)
    procedure Draw(const Canvas: TCanvas); override;
  end;

  TEbgalpha = class(TElement)
    procedure Update; override;
  end;

  TEcolor = class(TElement)
    Color: Integer;
    procedure Draw(const Canvas: TCanvas); override;
    procedure Update; override;
  end;

  TEdclock = class(TElement)
    Hour: Integer;
    Min: Integer;
    Sec: Integer;
    procedure Draw(const Canvas: TCanvas); override;
    procedure Update; override;
  end;

  TEfont = class(TElement)
    Font: TFont;
    procedure Draw(const Canvas: TCanvas); override;
    procedure Update; override;
  end;

  TElinegraph = class(TGraphElement)
    procedure Draw(const Canvas: TCanvas); override;
  end;

  TEspace = class(TElement)
    procedure Update; override;
  end;

  TEtext = class(TElement)
    CText: WideString;
    LText: WideString;
    RText: WideString;
    procedure Draw(const Canvas: TCanvas); override;
    procedure Update; override;
  end;

  TEwidth = class(TElement)
    procedure Update; override;
  end;

const
  ElementClasses: array[0..9] of TElementClass = (
    TEareagraph,
    TEbargraph,
    TEbgalpha,
    TEcolor,
    TEdclock,
    TEfont,
    TElinegraph,
    TEspace,
    TEtext,
    TEwidth
    );

var
  FontCache: array of TFont;
  FontCacheSize: Integer;

function SIntDef(const S: string): Integer;
begin
  try
    Result := SInt(S);
  except
    on EValueError do Result := 0;
  end;
end;

function SInt64Def(const S: string): Int64;
begin
  try
    Result := SInt64(S);
  except
    on EValueError do Result := 0;
  end;
end;

function SFloatDef(const S: string): Double;
begin
  try
    Result := SFloat(S);
  except
    on EValueError do Result := 0;
  end;
end;

function HIntDef(const S: string): Integer;
begin
  try
    Result := HInt(S);
  except
    on EValueError do Result := 0;
  end;
end;

function CreateElement(const Args: TStringArray): TElement;
var
  i: Integer;
  cond: string;
  params: TStringArray;
begin
  Result := nil;
  params := nil;

  if Length(Args) < 2 then Exit;

  if Args[0] = 'if' then begin
    cond := Args[1];
    params := SliceArray(Args, 2);
  end else begin
    cond := '';
    params := Args;
  end;

  if Length(params) < 2 then Exit;

  for i := 0 to High(ElementClasses) do
    if ElementClasses[i].ClassNameIs('TE' + params[0]) then begin
      Result := ElementClasses[i].Create(cond, params);
      Exit;
    end;
end;

procedure ClearFontCache;
var
  i: Integer;
begin
  for i := 0 to FontCacheSize-1 do
    FontCache[i].Free;
  SetLength(FontCache, 0);
  FontCacheSize := 0;
end;

function GetValue(const S: string): string;
begin
  if S[1] = '$' then
    Result := Monitor.GetValue(Copy(S, 2, Length(S)-1))
  else
    Result := S;
end;

function Fmt(const Args: array of string): string;
var
  i, j: Integer;
  pflg, wsign: Boolean;
  w, s, fmtstr: string;
  c: Char;
begin
  i := 0;
  j := 2;
  pflg := False;
  wsign := False;
  fmtstr := Args[1];
  Result := '';
  while i < Length(FmtStr) do
  begin
    Inc(i);
    if pflg then
    begin
      case FmtStr[i] of
        '%':
          begin
            Result := Result + '%';
            pflg := False;
          end;
        '-':
          wsign := True;
        '0'..'9':
          w := w + FmtStr[i];
        's', 'h':
          begin
            if Copy(w, 1, 1) = '0' then
              c := '0'
            else
              c := ' ';
            if FmtStr[i] = 'h' then
              s := UnitFormat(SInt64Def(Args[j]))
            else
              s := Args[j];
            if wsign then
              Result := Result + s + StringOfChar(c, SIntDef(w)-Length(s))
            else
              Result := Result + StringOfChar(c, SIntDef(w)-Length(s)) + s;
            Inc(j);
            pflg := False;
          end;
      end;
    end
    else if FmtStr[i] = '%' then
    begin
      pflg := True;
      w := '';
    end else
      Result := Result + FmtStr[i];
  end;
end;


{ TElement }

constructor TElement.Create(const Cond: string; const Args: TStringArray);
begin
  Self.Cond := Cond;
  Self.Args := Args;
  SetLength(Self.Values, Length(Self.Args));
end;

procedure TElement.Draw(const Canvas: TCanvas);
begin
  Inc(Current.Y, Self.Height);
end;

procedure TElement.DrawAlpha(const Canvas: TCanvas);
var
  c: Integer;
begin
  c := Current.Color;
  Current.Color := $ff;
  Draw(Canvas);
  Current.Color := c;
end;

procedure TElement.Init;
var
  i: Integer;
begin
  for i := 1 to High(Args) do
    if Args[i][1] = '$' then
      UseValue(Copy(Args[i], 2, Length(Args[i])-1))
end;

procedure TElement.Update;
var
  i: Integer;
begin
  for i := 1 to High(Args) do
    Values[i] := GetValue(Args[i]);

  if Cond = '' then
    CondValue := True
  else
    CondValue := GetValue(Cond) <> '';
end;


{ TGraphElement }

procedure TGraphElement.Update;
var
  i: Integer;
  d: Double;
begin
  inherited;
  Range := 0;
  if not Assigned(Data) then
    SetLength(Data, Width-2);
  for i := 1 to High(Values) do
    case i of
      1: Height := SIntDef(Values[i]);
      2: Range := SFloatDef(Values[i]);
      3: begin
           d := SFloatDef(Values[i]);
           Data[Cur] := (Prev + d) / 2;
           Prev := d;
         end;
    end;
  Inc(Cur);
  if Cur = Length(Data) then
    Cur := 0;
end;


{ areagraph }

procedure TEareagraph.Draw(const Canvas: TCanvas);
var
  i, h: Integer;
begin
  if Range <> 0 then
  begin
    for i := 0 to High(Data) do
    begin
      h := Trunc(Data[(i + Cur) mod Length(Data)] * Height / Range);
      if h > Height then
        h := Height;
      if h > 0 then
        Canvas.FillRect(1+i, Current.Y+Height-h, 1, h+1, Current.Color);
    end;
  end;
  inherited;
end;


{ bargraph }

procedure TEbargraph.Draw(const Canvas: TCanvas);
var
  i, w: Integer;
begin
  if Range <> 0 then
  begin
    w := Trunc(Data[(High(Data) + Cur) mod Length(Data)] * Width / Range);
    for i := 0 to w div 2 - 1 do
      Canvas.FillRect(1+i*2, Current.Y, 1, Self.Height, Current.Color);
  end;
  inherited;
end;


{ bgalpha }

procedure TEbgalpha.Update;
begin
  inherited;
  BgAlpha := SIntDef(Values[1]);
end;


{ color }

procedure TEcolor.Draw(const Canvas: TCanvas);
begin
  Current.Color := Color;
end;

procedure TEcolor.Update;
begin
  inherited;
  Color := HIntDef(Values[1]);
end;


{ dclock }

procedure TEdclock.Draw(const Canvas: TCanvas);
  var
    w, h: Integer;

  procedure DrawChar(X: Integer; C: Char);
  const
    {
       00
      3  4
       11
      5  6
       22
    }
    tab: array[0..9, 0..6] of Integer = (
      (1,0,1,1,1,1,1),
      (0,0,0,0,1,0,1),
      (1,1,1,0,1,1,0),
      (1,1,1,0,1,0,1),
      (0,1,0,1,1,0,1),
      (1,1,1,1,0,0,1),
      (1,1,1,1,0,1,1),
      (1,0,0,0,1,0,1),
      (1,1,1,1,1,1,1),
      (1,1,1,1,1,0,1)
    );
  var
    n: Integer;
  begin
    with Canvas do
      with Current do
        case c of
          ':': FillRect(X, Y+H div 2, 2, 2, color); // upper dot
          '.': FillRect(X, Y+H*2+1-H div 2, 2, 2, color); // lower dot
          '0'..'9':
            begin
              n := Ord(c) - Ord('0');
              if tab[n, 0] = 1 then FillRect(X+1,   Y,       W-2, 1, color);
              if tab[n, 1] = 1 then FillRect(X+1,   Y+H+1,   W-2, 1, color);
              if tab[n, 2] = 1 then FillRect(X+1,   Y+H*2+2, W-2, 1, color);
              if tab[n, 3] = 1 then FillRect(X,     Y+1,     2,   H, color);
              if tab[n, 4] = 1 then FillRect(X+W-2, Y+1,     2,   H, color);
              if tab[n, 5] = 1 then FillRect(X,     Y+H+2,   2,   H, color);
              if tab[n, 6] = 1 then FillRect(X+W-2, Y+H+2,   2,   H, color);
            end;
        end;
  end;
const
  hlineh = 1; // horizontal line height
  dotw = 2; // dot width
  x = 2;
var
  cx, sp, tw: Integer;
begin
  w := Trunc((Width-2) / 5 + 0.5); // "00:00" length is 5
  sp := Trunc(w / 6 + 0.5); // letter space
  w := w - sp; // digit width
  tw := w*4 + dotw+2 + sp*4; // total width
  h := (Height - hlineh*3) div 2;
  cx := (Width - tw) div 2;
  DrawChar(cx, Char(Ord('0') + Hour div 10));
  Inc(cx, w+sp);
  DrawChar(cx, Char(Ord('0') + Hour mod 10));
  Inc(cx, w+sp+1);
  if (Sec and 1) = 0 then
    DrawChar(cx, ':')
  else
    DrawChar(cx, '.');
  Inc(cx, dotw+1+sp);
  DrawChar(cx, Char(Ord('0') + Min div 10));
  Inc(cx, w+sp);
  DrawChar(cx, Char(Ord('0') + Min mod 10));
  inherited;
end;

procedure TEdclock.Update;
var
  i, v: Integer;
begin
  inherited;
  for i := 1 to High(Values) do
  begin
    v := SIntDef(Values[i]);
    case i of
      1: Height := v;
      2: Hour := v;
      3: Min := v;
      4: Sec := v;
    end;
  end;
end;


{ font }

procedure TEfont.Draw(const Canvas: TCanvas);
begin
  Canvas.SetFont(Font);
end;

procedure TEfont.Update;
var
  i: Integer;
  name: string;
  size: Integer;
  bold, italic: Boolean;
  quality: TFontQuality;
  f: TFont;
begin
  inherited;
  name := '';
  size := 0;
  bold := False;
  italic := False;
  quality := fqAntiAliased;
  for i := 1 to High(Values) do
    case i of
      1: name := DecodeUTF8(Values[i]);
      2: size := SIntDef(Values[i]);
      3, 4, 5:
        if Values[i] = 'bold' then
          bold := True
        else if Values[i] = 'italic' then
          italic := True
        else if Values[i] = 'cleartype' then
          quality := fqClearType;
    end;

  if Assigned(Font) and
      (Font.Name = name) and (Font.Size = size) and (Font.Bold = bold) and
      (Font.Italic = italic) and (Font.Quality = quality) then
    Exit;

  for i := 0 to FontCacheSize-1 do
  begin
    f := FontCache[i];
    if (f.Name = name) and (f.Size = size) and (f.Bold = bold) and
      (f.Italic = italic) and (f.Quality = quality) then
    begin
      Font := f;
      Exit;
    end;
  end;

  Font := TFont.Create(Name, Size, Bold, Italic, Quality);
  if Length(FontCache) = FontCacheSize then
    SetLength(FontCache, Length(FontCache) + 4);
  FontCache[FontCacheSize] := Font;
  Inc(FontCacheSize);
end;


{ linegraph }

procedure TElinegraph.Draw(const Canvas: TCanvas);
var
  i, h, prevh, x1, y1, x2, y2: Integer;
begin
  prevh := 0;
  if Range <> 0 then
  begin
    for i := 0 to High(Data) do
    begin
      h := Trunc(Data[(i + Cur) mod Length(Data)] * Height / Range);
      if h < 0 then
        h := 0
      else if h > Height then
        h := Height;
      if i = 0 then
        prevh := h;
      x1 := 1 + i;
      y1 := Current.Y + Height - prevh;
      x2 := x1 + 1;
      y2 := Current.Y + Height - h;
      if h = 0 then
      begin
        if prevh <> 0 then
          Canvas.Line(x1, y1, x2, y2+1, Current.Color);
      end else
        Canvas.Line(x1, y1, x2, y2, Current.Color);
      prevh := h;
    end;
  end;
  inherited;
end;


{ space }

procedure TEspace.Update;
begin
  inherited;
  Height := SIntDef(Values[1]);
end;


{ text }

procedure TEtext.Draw(const Canvas: TCanvas);
begin
  Canvas.SetTextColor(Current.Color);
  if CText <> '' then
  begin
    Canvas.SetTextAlign(taCenter);
    Canvas.DrawText(Width div 2, Current.Y, CText);
  end
  else begin
    Canvas.SetTextAlign(taLeft);
    Canvas.DrawText(1, Current.Y, LText);
    Canvas.SetTextAlign(taRight);
    Canvas.DrawText(Width, Current.Y, RText);
  end;
  Height := Canvas.TextMetrics.Height;
  inherited;
end;


{ width }

procedure TEwidth.Update;
begin
  inherited;
  Width := SIntDef(Values[1]);
end;

procedure TEtext.Update;
var
  i, j: Integer;
  s, t: string;
  esc: Boolean;
begin
  inherited;
  s := Values[1];
  esc := False;
  for i := 1 to Length(s) do
    case s[i] of
      '\': if esc then
           begin
             t := t + '\';
             esc := False;
           end else
             esc := True;
      's': if esc then
           begin
             t := t + #0;
             esc := False;
           end else
             t := t + 's';
    else
      if esc then
      begin
        t := t + '\';
        esc := False;
      end;
      t := t + s[i];
    end;
  Values[1] := t;
  s := Fmt(Values);
  i := 0;
  for j := 1 to Length(s) do
    if s[j] = #0 then
    begin
      i := j;
      Break;
    end;

  if i = 0 then
    CText := DecodeUTF8(s)
  else begin
    CText := '';
    LText := DecodeUTF8(Copy(s, 1, i-1));
    RText := DecodeUTF8(Copy(s, i, Length(s)-i+1));
  end;
end;

end.
