unit NNet;

{$IMPORTEDDATA OFF}

interface

uses Windows, WinAdd, Winsock2Sub, Iphlpapi, NLib;

type
  EIpError = class(Exception)
    constructor Create(ErrorCode: IP_STATUS);
  end;

  // SOCKADDR_IN and SOCKADDR_IN6
  TAddress = record
    Family: ADDRESS_FAMILY;
    Port: USHORT;
    case Integer of
      0: (Addr: IN_ADDR; Zero: array [0..7] of Char);
      1: (FlowInfo: ULONG; Addr6: IN6_ADDR; ScopeId: ULONG);
  end;
  PAddress = ^TAddress;

  TUdp = class
  private
    Soc: SOCKET_;
    Family: ADDRESS_FAMILY;
    Address: TAddress;
    AddressLen: Integer;
    Overlapped: WSAOVERLAPPED;
    ReadBuf: WSABUF;
    procedure DoRecv;
  public
    OnRecv: procedure (NRead: Cardinal; const Address: TAddress) of object;
    constructor Create(Family: ADDRESS_FAMILY = AF_INET);
    destructor Destroy; override;
    procedure Bind(Port: Cardinal);
    function Send(const Buf; NBytes: Integer; const Address: TAddress): Integer;
    function Recv(var Buf; NBytes: Integer; var Address: TAddress): Integer;
    procedure RecvAsync(var Buf; NBytes: Integer);
  end;

  TIcmpEchoReply = record
    Address: TAddress;
    Ttl: Integer;
    Rtt: Integer;
  end;

  TIcmp = class
    Handle: THandle;
    constructor Create(Family: ADDRESS_FAMILY = AF_INET);
    destructor Destroy; override;
    function SendEcho(const Data; NBytes: Integer; const Address: TAddress): TIcmpEchoReply;
  end;

function FormatAddress(const Address: TAddress): WideString; overload;
function FormatAddress(Address: Cardinal): WideString; overload;
function GetAddress(const Host: WideString; Port: Cardinal = 0): TAddress;
function GetHostName: WideString;
function GetName(Address: Cardinal): WideString;

implementation

uses NSys;

constructor EIpError.Create(ErrorCode: IP_STATUS);
var
  buf: array[0..1023] of WideChar;
  buflen: DWORD;
begin
  buflen := Length(buf);
  if GetIpErrorString(ErrorCode, @buf, @buflen) = NO_ERROR then
    Msg := buf
  else
    Msg := 'Error ' + IStr(ErrorCode);
end;

function FormatAddress(const Address: TAddress): WideString;
var
  a: TAddress;
  buf: array [0..INET6_ADDRSTRLEN-1] of WideChar;
  buflen: DWORD;
begin
  a := Address;
  a.Port := 0;
  buflen := Length(buf);
  if WSAAddressToStringW(@a, SizeOf(a), nil, @buf, @buflen) <> 0 then
    raise EOSError.Create(WSAGetLastError);
  Result := buf;
end;

function FormatAddress(Address: Cardinal): WideString;
var
  a: TAddress;
begin
  a.Family := AF_INET;
  a.Addr.s_addr := Address;
  Result := FormatAddress(a);
end;

// XP SP2 and later (GetAddrInfoW)
function GetAddress(const Host: WideString; Port: Cardinal = 0): TAddress;
var
  pa: PADDRINFOW;
begin
  if GetAddrInfoW(PWideChar(Host), nil, nil, pa) <> 0 then
    raise EOSError.Create(WSAGetLastError);
  Result := PAddress(pa.ai_addr)^;
  Result.Port := HtoBe16(Port);
  FreeAddrInfoW(pa);
end;

function GetHostName: WideString;
var
  buf: array [0..255] of Char;
begin
  if Winsock2Sub.gethostname(@buf, SizeOf(buf)) <> 0 then
    raise EOSError.Create(WSAGetLastError);
  Result := buf
end;

// XP SP2 and later (GetNameInfoW)
function GetName(Address: Cardinal): WideString;
var
  sa: SOCKADDR_IN;
  buf: array [0..NI_MAXHOST-1] of WideChar;
begin
  FillChar(sa, SizeOf(sa), 0);
  sa.sin_family := AF_INET;
  sa.sin_port := 0;
  sa.sin_addr := IN_ADDR(Address);
  if GetNameInfoW(@sa, SizeOf(sa), @buf, Length(buf), nil, 0, 0) <> 0 then
    raise EOSError.Create(WSAGetLastError);
  Result := buf;
end;

{ TUdp }

constructor TUdp.Create(Family: ADDRESS_FAMILY = AF_INET);
begin
  Self.Family := Family;
  Soc := socket(Family, SOCK_DGRAM, IPPROTO_UDP);
  if Soc = INVALID_SOCKET then
    raise EOSError.Create(WSAGetLastError);
end;

destructor TUdp.Destroy;
begin
  closesocket(Soc);
end;

procedure TUdp.Bind(Port: Cardinal);
var
  a: TAddress;
begin
  FillChar(a, SizeOf(a), 0);
  a.Family := Family;
  a.Port := HToBe16(Port);
  if Winsock2Sub.bind(Soc, @a, SizeOf(a)) = SOCKET_ERROR then
    raise EOSError.Create(WSAGetLastError);
end;

function TUdp.Send(const Buf; NBytes: Integer; const Address: TAddress): Integer;
begin
  Result:= sendto(Soc, @Buf, NBytes, 0, @Address, SizeOf(Address));
  if Result = SOCKET_ERROR then
    raise EOSError.Create(WSAGetLastError);
end;

function TUdp.Recv(var Buf; NBytes: Integer; var Address: TAddress): Integer;
var
  salen: Integer;
begin
  salen := SizeOf(Address);
  Result := recvfrom(Soc, @Buf, NBytes, 0, @Address, @salen);
  if Result = SOCKET_ERROR then
    raise EOSError.Create(WSAGetLastError);
end;

procedure UdpRecvAsyncDone(dwError: DWORD; cbTransferred: DWORD;
    lpOverlapped: LPWSAOVERLAPPED; dwFlags: DWORD); stdcall;
begin
  if dwError = ERROR_OPERATION_ABORTED then
    Exit;
  with TUdp(lpOverlapped.hEvent) do begin
    OnRecv(cbTransferred, Address);
    DoRecv;
  end;
end;

procedure TUdp.RecvAsync(var Buf; NBytes: Integer);
begin
  ReadBuf.buf := @Buf;
  ReadBuf.len := NBytes;
  AddressLen := SizeOf(Address);
  DoRecv;
end;

procedure TUdp.DoRecv;
var
  flags: DWORD;
begin
  flags := 0;
  FillChar(Overlapped, SizeOf(Overlapped), 0);
  Overlapped.hEvent := THandle(Self);
  if WSARecvFrom(Soc, @ReadBuf, 1, nil, @flags, @Address, @AddressLen,
      @Overlapped, UdpRecvAsyncDone) = SOCKET_ERROR then
    if WSAGetLastError <> WSA_IO_PENDING then
      raise EOSError.Create(WSAGetLastError);
end;

{ TIcmp }

constructor TIcmp.Create(Family: ADDRESS_FAMILY = AF_INET);
begin
  case Family of
    AF_INET: Handle := IcmpCreateFile();
    AF_INET6: Handle := Icmp6CreateFile();
  else
    raise EValueError.Create;
  end;
  if Handle = INVALID_HANDLE_VALUE then
    raise EOSError.Create;
end;

destructor TIcmp.Destroy;
begin
  IcmpCloseHandle(Handle);
end;

function TIcmp.SendEcho(const Data; NBytes: Integer; const Address: TAddress): TIcmpEchoReply;
var
  r: PIP_ECHO_REPLY;
  r6: PICMPV6_ECHO_REPLY;
  rsize: Integer;
  saddr6: SOCKADDR_IN6;
  options: IP_OPTION_INFORMATION;
begin
  case Address.Family of
    AF_INET:
      begin
        rsize := SizeOf(IP_ECHO_REPLY) + NBytes;
        GetMem(r, rsize);
        try
          FillChar(r^, SizeOf(rsize), 0);
          if IcmpSendEcho(Handle, Address.Addr.S_addr, @Data, NBytes, nil, r,
              rsize, $ffffffff) = 0 then
            raise EOSError.Create;
          if r.Status <> 0 then
            raise EIpError.Create(r.Status);
          Result.Address.Addr.S_addr := r.Address;
          Result.Ttl := Ord(r.Options.Ttl);
          Result.Rtt := r.RoundTripTime;
        finally
          FreeMem(r);
        end;
      end;
    AF_INET6:
      begin
        rsize := SizeOf(ICMPV6_ECHO_REPLY) + NBytes;
        GetMem(r6, rsize);
        try
          FillChar(r6^, SizeOf(rsize), 0);
          FillChar(saddr6, SizeOf(saddr6), 0);
          saddr6.sin6_family := AF_INET6;
          // options is not optional on Windows Server 2003 and XP
          FillChar(options, SizeOf(options), 0);
          options.Ttl := Char(255);
          if Icmp6SendEcho2(Handle, 0, nil, nil, @saddr6, @Address, @Data, NBytes,
             @options, r6, rsize, $ffffffff) = 0 then
            raise EOSError.Create;
          if r6.Status <> 0 then
            raise EIpError.Create(r6.Status);
          Move(r6.Address.sin6_addr, Result.Address.Addr6, SizeOf(Result.Address.Addr6));
          Result.Address.ScopeId := r6.Address.sin6_scope_id;
          Result.Ttl := -1;
          Result.Rtt := r6.RoundTripTime;
        finally
          FreeMem(r6);
        end;
      end;
  else
    raise EValueError.Create;
  end;
  Result.Address.Family := Address.Family;
end;

var
  wsadata_: WSADATA;

initialization
  WSAStartUp($0202, @wsadata_);

finalization
  WSACleanUp;

end.
